import logging
from collections import Counter
from typing import Dict, List

import chromadb

from factflip.llms import Llm, get_default_llm
from factflip.utils import dequote
from factflip.validators import DefaultMemeValidator, MemeValidator


class FactFlipMemeGenerator(object):
    def __init__(
        self,
        embeddings: chromadb.Collection,
        llm: Llm = None,
    ) -> None:
        self.embeddings = embeddings

        if not llm:
            self.llm = get_default_llm()
        else:
            self.llm = llm

    @staticmethod
    def _generate_messages(claim: str, review: str, examples) -> List[dict]:
        messages = [
            {
                "role": "system",
                "content": "You are a fact-checker and meme expert that provides short and acurate answers.",
            }
        ]

        for e in examples:
            title = e["imgflip_template"].split("/")[-1].replace("-", " ").strip()

            messages.append(
                {
                    "role": "user",
                    "content": f"Instruction: Generate a caption to turn the image into a meme. The meme must contains {e['kym_template_usage_concept']} and humour and be formulated as response to a erronous claim based on the provided claim review.",  # noqa E501
                }
            )
            messages.append(
                {
                    "role": "user",
                    "content": f"Input image: The image is \"{title}\". {e['imgflip_template_image_description']} {e['kym_template_usage']}",  # noqa E501
                }
            )
            messages.append(
                {
                    "role": "user",
                    "content": f"Input erronous claim: {e['imgflip_instance_claim']}",
                }
            )
            messages.append(
                {
                    "role": "user",
                    "content": f"Input claim review: {e['imgflip_instance_correction_claim']}",
                }
            )
            messages.append(
                {
                    "role": "assistant",
                    "content": f"Output meme caption: {e['imgflip_instance_text']}",
                }
            )

        messages.append(
            {
                "role": "user",
                "content": f"Instruction: Generate a caption to turn the image into a meme. The meme must contains {examples[0]['kym_template_usage_concept']} and humour and be formulated as response to a erronous claim based on the provided claim review.",  # noqa E501
            }
        )
        messages.append(
            {
                "role": "user",
                "content": f"Input image: The image is \"{title}\". {examples[0]['imgflip_template_image_description']} {examples[0]['kym_template_usage']}",  # noqa E501
            }
        )
        messages.append({"role": "user", "content": f"Input erronous claim: {claim}"})
        messages.append({"role": "user", "content": f"Input claim review: {review}"})

        return messages

    def generate_meme_text(self, claim: str, review: str, validator: MemeValidator = None) -> Dict[str, str]:
        log = logging.getLogger()
        if not validator:
            validator = DefaultMemeValidator(1)

        # 1) Find meme candidates and get the most comon template:
        candidates = self.embeddings.query(query_texts=[claim], n_results=10)
        meme_template = Counter([x["imgflip_template"] for x in candidates["metadatas"][0]]).most_common(1)[0][0]

        # 2) Find examples for the meme template:
        examples = self.embeddings.query(
            query_texts=["This is a question or text"],
            n_results=4,  # Best number found by memecraft.
            where={"imgflip_template": meme_template},
        )
        meme_template_id = examples["metadatas"][0][0]["imgflip_template_id"]

        # 3) Generate the prompt usig the examples
        messages = self.__class__._generate_messages(claim=claim, review=review, examples=examples["metadatas"][0])

        # 4) Generate the the meme:
        for i in range(max(1, validator.max_retries)):
            response = self.llm.chat(messages)

            # 5) Split the sentence in two:
            split_response = self.llm.generate(
                prompt=f"Just answer: Split the sentence in two: {response}",
            ).splitlines()[-2:]

            split_response = [dequote(x) for x in split_response]  # Remove quotes if present.

            if validator.validate_caption(split_response):
                log.info(split_response)
                break
            else:
                log.info(f"{i}/{max(1, validator.max_retries)} The meme was not valid: {split_response}")

        return {
            "template_url": meme_template,
            "template_id": meme_template_id,
            "meme_caption": split_response,
            "meme_claim": claim,
            "meme_claim_review": review,
        }



import os
from pathlib import Path

import chromadb
import pandas as pd
import requests
from imgflip import AsyncMeme, Imgflip, SyncMeme
from rich.progress import track

from factflip.generators import FactFlipMemeGenerator
from factflip.llms import OllamaLlm
from factflip.utils import dequote
from factflip.validators import FactFlipMemeValidator


def create_memes_from_summaries(embeddings: chromadb.Collection, fc_df: pd.DataFrame, drop_na=False) -> pd.DataFrame:
    model = "llava:13b-v1.6"
    generator = FactFlipMemeGenerator(embeddings=embeddings, llm=OllamaLlm(model))

    if drop_na:
        memes = fc_df[fc_df.verdict.notna()].progress_apply(
            lambda row: generator.generate_meme_text(
                claim=row.claim, review=row.verdict, validator=FactFlipMemeValidator(10)
            ),
            axis=1,
        )
    else:
        memes = fc_df.progress_apply(
            lambda row: generator.generate_meme_text(
                claim=row.claim, review=row.verdict, validator=FactFlipMemeValidator(10)
            ),
            axis=1,
        )
    return pd.DataFrame(memes.to_list())


def create_imgflip_meme(meme: dict, imgflip_username: str, imgflip_password: str) -> SyncMeme | AsyncMeme:
    imgflip = Imgflip(username=imgflip_username, password=imgflip_password, session=requests.Session())

    if len(meme["meme_caption"]) > 1:
        imgflipmeme = imgflip.make_meme(
            template=meme["template_id"],
            top_text=meme["meme_caption"][0],
            bottom_text=meme["meme_caption"][1],
        )
    else:
        imgflipmeme = imgflip.make_meme(template=meme["template_id"], top_text=meme["meme_caption"][0])
    return imgflipmeme


def create_imgflip_memes(generated_memes_df: pd.DataFrame, output_dir: Path):
    images = []
    for _, row in track(generated_memes_df.iterrows(), description="Creating Imgflip memes..."):
        img_meme = create_imgflip_meme(
            {
                "template_id": row.template_id,
                "meme_caption": [dequote(x) for x in row.meme_caption],
            }
        )
        img_filename = img_meme.url.split("/")[-1]
        img_meme.save(os.path.join(output_dir, img_filename))

        images.append(
            {
                "meme_imgage_page_url": img_meme.page_url,
                "meme_imgage_url": img_meme.url,
                "meme_image_filename": img_filename,
            }
        )

    return pd.DataFrame(images)



import abc
from typing import Dict, Sequence

import ollama

DEFAULT_OLLAMA_LMM = "llama3"


def get_default_ollama_llm() -> str:
    return DEFAULT_OLLAMA_LMM


class Llm(abc.ABC):
    @abc.abstractmethod
    def generate(self, prompt: str) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def chat(self, messages: Sequence[Dict[str, str]]) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def embed(self, prompt: str) -> Sequence[float]:
        raise NotImplementedError


class OllamaLlm(Llm):
    def __init__(self, model: str = None) -> None:
        super().__init__()
        if not model:
            self.model = get_default_ollama_llm()
        else:
            self.model = model

    def generate(self, prompt: str) -> str:
        return ollama.generate(model=self.model, prompt=prompt)["response"].strip()

    def chat(self, messages: Sequence[Dict[str, str]]) -> str:
        return ollama.chat(model=self.model, messages=messages)["message"]["content"].strip()

    def embed(self, prompt: str) -> Sequence[float]:
        return ollama.embeddings(model=self.model, prompt=prompt)


def get_default_llm() -> Llm:
    return OllamaLlm()



import os
import shutil
from pathlib import Path
from typing import Optional

import chromadb
import pandas as pd
import typer
from dotenv import load_dotenv
from rich.console import Console
from rich.pretty import pprint
from typing_extensions import Annotated

import factflip
from factflip import DEFAULT_EMBEDDINGS_PATH, DEFAULT_ROOT_PATH
from factflip.generators import FactFlipMemeGenerator
from factflip.llms import OllamaLlm
from factflip.utils.build import (
    build_embeddings_database,
    describe_imkg_memes,
    generate_imkg_claims,
    merge_claims_data,
    query_imkg_memes,
)
from factflip.utils.data import (
    download_and_parse_imkg,
    download_factflip_embeddings,
    download_imgflip_templates_images,
    load_rdf,
)
from factflip.utils.memes import create_imgflip_meme
from factflip.validators import FactFlipMemeValidator

load_dotenv()
app = typer.Typer(no_args_is_help=True, add_completion=False)


def _version_callback(value: bool):
    if value:
        print(factflip.version)
        raise typer.Exit()


@app.callback()
def callback(
    version: Annotated[
        Optional[bool],
        typer.Option("--version", "-v", help="Show the installed factflip version.", callback=_version_callback),
    ] = None,
):
    """🖼️  Factflip - A claim-based meme generator powered by IMKG and LLMs."""


@app.command(help="Build the meme embedding database.")
def build(
    imkg_dir: Annotated[
        Path,
        typer.Option(
            ...,
            "--imkg-dir",
            "-imkg",
            envvar="IMKG_DIR",
            help="The path where IMGK is stored.",
        ),
    ] = DEFAULT_ROOT_PATH,
    embeddings_db: Annotated[
        Path,
        typer.Option(
            ...,
            "--embeddings-db",
            "-db",
            envvar="FACTFLIP_EMBEDDINGS",
            help="The path of the FactFlip database that contains the embeddings.",
        ),
    ] = DEFAULT_EMBEDDINGS_PATH,
    download_imkg: Annotated[
        bool,
        typer.Option(
            "--download-imkg",
            "-d",
            help="Download IMKG release (this will not use the already downloaded and parsed IMKG data).",
        ),
    ] = False,
):
    Path(imkg_dir).mkdir(parents=True, exist_ok=True)
    templates_dir = os.path.join(imkg_dir, "templates")
    imkgf = os.path.join(imkg_dir, "imkg_full.ttl")
    imkg_memes_f = os.path.join(imkg_dir, "imkg_memes.csv")
    imkg_memes_descriptions_f = os.path.join(imkg_dir, "imkg_memes_descriptions.csv")
    imkg_claims_f = os.path.join(imkg_dir, "imkg_claims.csv")
    merged_claims_f = os.path.join(imkg_dir, "merged_imkg_claims.csv")

    console = Console()
    with console.status("Building embedding data (this will take a very long time)...") as status:
        # 1) Download and parse IMKG into one Graph.
        g = None
        if download_imkg or not os.path.isfile(imkgf):
            status.update("Downloading and parsing IMKG (this will take a very long time)...")
            g = download_and_parse_imkg()

            status.update("Saving parsed IMKG knowledge graph (this will take a very long time)...")
            g.serialize(destination=imkgf)

            status.update(f"IMKG knowledge graph saved: {imkgf}.")

        # 2) Download IMKG images.
        if os.path.isfile(imkg_memes_f):
            imkg_memes_df = pd.read_csv(imkg_memes_f)
        else:
            if not download_imkg and not g:
                status.update(f"Loading IMKG: {imkgf}.")
                g = load_rdf(imkgf)
                status.update("IMKG loaded.")

            status.update("Querying template images...")
            imkg_memes_df = query_imkg_memes(g)
            imkg_memes_df.to_csv(imkg_memes_f, index=False)

            status.update(f"Template query data saved: {imkg_memes_f}")

            status.update("Downloading template images from ImgFlip...")
            download_imgflip_templates_images(
                imkg_memes_df.imgflip_template.unique(),
                output_dir=templates_dir,
            )
            status.update("Template images saved.")

        # 3) Create images meme descriptions:
        if os.path.isfile(imkg_memes_descriptions_f):
            imkg_memes_descriptions_df = pd.read_csv(imkg_memes_descriptions_f)
        else:
            status.update("Generating memes descriptions...")
            imkg_memes_descriptions_df = describe_imkg_memes(
                imkg_memes_df,
                imgflip_templates_dir=templates_dir,
            )
            imkg_memes_descriptions_df.to_csv(imkg_memes_descriptions_f, index=False)
            status.update(f"Memes description data saved: {imkg_memes_descriptions_f}")

        # 4) Generate claims for existing memes:
        if os.path.isfile(imkg_claims_f):
            imkg_claims_df = pd.read_csv(imkg_claims_f)
        else:
            status.update("Generating claims...")
            imkg_claims_df = generate_imkg_claims(
                imkg_memes_df=imkg_memes_df, imkg_memes_descriptions_df=imkg_memes_descriptions_df, llm=OllamaLlm()
            )
            imkg_claims_df.to_csv(imkg_claims_f, index=False)
            status.update(f"Claims data saved: {imkg_claims_f}")

        # 5) Create embeddings database:
        status.update("Bulding embeddings database...")
        merged_claims_df = merge_claims_data(imkg_memes_df, imkg_memes_descriptions_df, imkg_claims_df)

        merged_claims_df.to_csv(merged_claims_f, index=False)
        status.update(f"Merged claims data saved: {merged_claims_f}")

        build_embeddings_database(merged_claims_df, embeddings_db, delete_collection=False)
        status.update("Embeddings database created.")


@app.command(help="Download precomputed Factflip embeddings.")
def download(
    embeddings_db: Annotated[
        Path,
        typer.Option(
            ...,
            "--embeddings-db",
            "-db",
            envvar="FACTFLIP_EMBEDDINGS",
            help="The path of the FactFlip database that contains the embeddings.",
        ),
    ] = DEFAULT_EMBEDDINGS_PATH,
    force_download: Annotated[
        bool,
        typer.Option("--force-download", "-f", help="Force download if the database is already present."),
    ] = False,
):
    console = Console()
    with console.status("Downloading embedding data...") as status:
        if not os.path.isdir(embeddings_db) or force_download:
            zipfile = download_factflip_embeddings(embeddings_version=factflip.version)

            status.update("Extracting embedding data...")
            if os.path.isdir(embeddings_db):
                shutil.rmtree(embeddings_db)

            Path(embeddings_db).mkdir(parents=True, exist_ok=True)
            zipfile.extractall(embeddings_db)
            status.update("Embedding data installed.")
        else:
            status.update("Embedding data already installed.")


@app.command(help="Render a meme instance using the imgflip API.")
def render(
    imgflip_template_id: Annotated[
        int,
        typer.Option(
            ...,
            "--imgflip-template-id",
            "-i",
            help="The ImgFlip template ID used for generating the final image.",
        ),
    ],
    caption_1: Annotated[
        str,
        typer.Argument(help="The first caption that is used for rendering the meme image."),
    ],
    caption_2: Annotated[
        str,
        typer.Argument(help="The optional second caption that is used for rendering the meme image."),
    ] = None,
    download_render: Annotated[
        bool,
        typer.Option("--download-render", "-d", help="Download rendered meme using ImgFlip API."),
    ] = False,
    imgflip_username: Annotated[
        str,
        typer.Option(
            ...,
            "--imgflip-username",
            "-u",
            envvar="IMGFLIP_USERNAME",
            help="The ImgFlip API username used for generating the final image.",
        ),
    ] = None,
    imgflip_password: Annotated[
        str,
        typer.Option(
            ...,
            "--imgflip-password",
            "-p",
            envvar="IMGFLIP_PASSWORD",
            help="The ImgFlip API password used for generating the final image.",
        ),
    ] = None,
):
    generated = {
        "template_id": imgflip_template_id,
        "meme_caption": [caption_1, caption_2] if caption_2 else [caption_1],
    }

    console = Console()
    with console.status("Rendering meme...") as status:
        if imgflip_username and imgflip_password:
            status.update("Generating meme image using FactFlip API...")
            imgflipmeme = create_imgflip_meme(
                meme=generated,
                imgflip_username=imgflip_username,
                imgflip_password=imgflip_password,
            )

            img_filename = imgflipmeme.url.split("/")[-1]
            generated["meme_imgage_page_url"] = imgflipmeme.page_url
            generated["meme_imgage_url"] = imgflipmeme.url
            generated["meme_image_filename"] = img_filename

            console.log("Meme image generated.")

            if download_render:
                status.update("Downloading meme image.")
                imgflipmeme.save(img_filename)
                status.update(f"Meme image saved: {img_filename}")

    pprint(generated, expand_all=True)


@app.command(help="Generate the text of a meme.")
def generate(
    claim: Annotated[
        str,
        typer.Argument(help="The claim (misinformation) that is used for generating the meme."),
    ],
    review: Annotated[
        str,
        typer.Argument(help="The claim review (claim correction) that is used for generating the meme."),
    ],
    embeddings_db: Annotated[
        Path,
        typer.Option(
            ...,
            "--embeddings-db",
            "-db",
            envvar="FACTFLIP_EMBEDDINGS",
            help="The path of the FactFlip database that contains the embeddings.",
        ),
    ] = DEFAULT_EMBEDDINGS_PATH,
    render: Annotated[
        bool,
        typer.Option("--render", "-r", help="Render the generated meme using ImgFlip API."),
    ] = False,
    download_render: Annotated[
        bool,
        typer.Option("--download-render", "-d", help="Download rendered meme using ImgFlip API."),
    ] = False,
    imgflip_username: Annotated[
        str,
        typer.Option(
            ...,
            "--imgflip-username",
            "-u",
            envvar="IMGFLIP_USERNAME",
            help="The ImgFlip API username used for generating the final image.",
        ),
    ] = None,
    imgflip_password: Annotated[
        str,
        typer.Option(
            ...,
            "--imgflip-password",
            "-p",
            envvar="IMGFLIP_PASSWORD",
            help="The ImgFlip API password used for generating the final image.",
        ),
    ] = None,
):
    console = Console()
    with console.status("Generating meme...") as status:
        status.update("Loading embeddings...")
        client = chromadb.PersistentClient(path=embeddings_db.absolute().as_posix())
        collection = client.get_or_create_collection("claims")
        console.log("Embeddings database loaded.")

        status.update("Loading language model...")
        generator = FactFlipMemeGenerator(embeddings=collection)
        console.log("Language model loaded.")

        status.update("Generating meme content...")
        generated = generator.generate_meme_text(claim, review, validator=FactFlipMemeValidator(max_retries=5))
        console.log("Meme content generated.")

        if (render or download_render) and imgflip_username and imgflip_password:
            status.update("Generating meme image using FactFlip API...")
            imgflipmeme = create_imgflip_meme(
                meme=generated,
                imgflip_username=imgflip_username,
                imgflip_password=imgflip_password,
            )

            img_filename = imgflipmeme.url.split("/")[-1]
            generated["meme_imgage_page_url"] = imgflipmeme.page_url
            generated["meme_imgage_url"] = imgflipmeme.url
            generated["meme_image_filename"] = img_filename

            console.log("Meme image generated.")

            if download_render:
                status.update("Downloading meme image.")
                imgflipmeme.save(img_filename)
                status.update(f"Meme image saved: {img_filename}")

    pprint(generated, expand_all=True)


if __name__ == "__main__":
    app()



import abc
from typing import List


class MemeValidator(abc.ABC):
    def __init__(self, max_retries: int = 5) -> None:
        super().__init__()
        self.max_retries = max_retries

    @abc.abstractmethod
    def validate_caption(self, captions: List[str]) -> bool:
        raise NotImplementedError


class DefaultMemeValidator(MemeValidator):
    def validate_caption(self, captions: List[str]) -> bool:
        return True


class FactFlipMemeValidator(MemeValidator):
    def validate_caption(self, captions: List[str]) -> bool:
        txt = " ".join(captions)

        if not txt.isupper():  # Consider relaxing that rule...
            return False
        if len(txt) < 10:
            return False
        if "1." in txt:
            return False
        if "2." in txt:
            return False
        if "meme" in txt:
            return False
        if "sorry" in txt:
            return False
        if "ASSISTANT" in txt:
            return False
        if "ASSIST" in txt:
            return False
        if "SPLIT SENTENCE" in txt:
            return False

        for i in captions:
            if len(i) <= 5:
                return False

        return True



import logging
import os
import urllib
import zipfile
from io import BytesIO
from pathlib import Path
from typing import List

import requests
from rdflib import Graph
from rich.progress import track

import factflip
from factflip import APP_REPOSITORY, DEFAULT_EMBEDDINGS_PATH, DEFAULT_TEMPLATES_PATH


def load_rdf(rdf: Path, format="ttl") -> Graph:
    g = Graph()
    g.parse(rdf, format=format)

    return g


def load_imkg(imkg_full_dir: str) -> Graph:
    g = Graph()
    for subdir, _, files in os.walk(imkg_full_dir):
        for file in files:
            ext = os.path.splitext(file)[-1].lower()
            if ext in [".nt", ".ttl"]:
                print(os.path.join(subdir, file))
                g.parse(os.path.join(subdir, file), format=ext[1::])

    return g


def download_and_parse_imkg(
    imkg_url: str = "https://zenodo.org/records/7457166/files/imkg.1.0.data.zip?download=1",
) -> Graph:
    log = logging.getLogger()
    tmpf, _ = urllib.request.urlretrieve(imkg_url)
    with zipfile.ZipFile(tmpf) as zf:
        g = Graph()
        log.info("Parsing: full/imgflip.nt.")
        g.parse(zf.read("full/imgflip.nt"), format="nt")

        for file in zf.namelist():
            if file.startswith("full/"):
                ext = os.path.splitext(file)[-1].lower()
                if ext in [".nt", ".ttl"]:
                    if file != "full/imgflip.nt":
                        log.info(f"Parsing: {file}.")
                        g.parse(zf.read(file), format=ext[1::])

    urllib.request.urlcleanup()
    return g


def download_imgflip_templates_images(
    imgflip_template: List[int] = None,
    output_dir: Path = DEFAULT_TEMPLATES_PATH,
    force_download: bool = False,
):
    if imgflip_template is None:
        imgflip_template = []

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.106 Safari/537.36"  # noqa E501
    }
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    for tmpl in track(imgflip_template, description="Downloading image templates..."):
        tmpl_name = tmpl.split("/")[-1]
        imgf = os.path.join(output_dir, f"{tmpl_name}.jpg")
        url = f"https://imgflip.com/s/meme/{tmpl_name}.jpg"

        if not os.path.isfile(imgf) or force_download:
            response = requests.get(url, stream=True, headers=headers)
            with open(imgf, "bw") as file:
                file.write(response.content)


def download_factflip_embeddings(
    repository_url: str = APP_REPOSITORY, embeddings_version: str = factflip.version
) -> zipfile.ZipFile:
    embedding_asset_url = urllib.parse.urljoin(
        f"{repository_url}/", f"releases/download/v{embeddings_version}/factflip_embeddings.zip"
    )
    request = requests.get(embedding_asset_url, stream=True)
    return zipfile.ZipFile(BytesIO(request.content))


def zip_factflip_embeddings(
    embeddings_db: Path = DEFAULT_EMBEDDINGS_PATH, output: Path = "./data/factflip_embeddings.zip"
):
    zipf = zipfile.ZipFile(output, "w", zipfile.ZIP_DEFLATED)

    for dirpath, _, filenames in os.walk(embeddings_db):
        for filename in filenames:
            if filename not in [".DS_Store", "__MACOSX"]:
                filepath = os.path.join(dirpath, filename)
                arcname = os.path.relpath(filepath, embeddings_db)
                zipf.write(filepath, arcname)
    zipf.close()



import os

from platformdirs import user_data_path
from usingversion import getattr_with_version

APPNAME = "factflip"
APP_REPOSITORY = "https://github.com/evhart/factflip"
DEFAULT_ROOT_PATH = user_data_path(appname=APPNAME)
DEFAULT_EMBEDDINGS_PATH = os.path.join(DEFAULT_ROOT_PATH, "factflip_embeddings")
DEFAULT_TEMPLATES_PATH = os.path.join(DEFAULT_ROOT_PATH, "templates")

__getattr__ = getattr_with_version(APPNAME, __file__, __name__)



import os
import re
from pathlib import Path
from typing import Sequence

import chromadb
import pandas as pd
from rdflib import Graph
from rich.progress import track

from factflip import DEFAULT_EMBEDDINGS_PATH, DEFAULT_TEMPLATES_PATH
from factflip.llms import Llm, get_default_llm


def describe_imkg_memes(
    imkg_memes_df: pd.DataFrame,
    imgflip_templates_dir: Path = DEFAULT_TEMPLATES_PATH,
    llm: Llm = None,
) -> pd.DataFrame:
    data = []
    if not llm:
        llm = get_default_llm()

    tmpl_df = imkg_memes_df.groupby(["kym_template", "imgflip_template", "usage"]).first().reset_index()
    for _, row in track(
        tmpl_df.iterrows(),
        total=len(tmpl_df),
        description="Generating memes descriptions...",
    ):
        tmpl = row["imgflip_template"].split("/")[-1]
        tmpl_file = os.path.join(imgflip_templates_dir, f"{tmpl}.jpg")

        concept = llm.chat(
            messages=[
                {
                    "role": "system",
                    "content": "You are a fact-checker and meme expert that provides short and acurate answers.",
                },
                {
                    "role": "user",
                    "content": "Just answer in one or two words without punctuation, what is the main rethorical device used in the following meme description?",  # noqa E501
                },
                {"role": "user", "content": row["usage"]},
            ],
        ).lower()

        desc = llm.chat(
            messages=[
                {
                    "role": "system",
                    "content": "You are a fact-checker and meme expert who perfectly describes images concisely.",
                },
                {
                    "role": "user",
                    "content": f"Provide a short description for the image. Consider that the image is an image used for {concept} and humor.",  # noqa E501
                    "images": [tmpl_file],
                },
            ],
        )

        data.append(
            {
                "imgflip_template": row["imgflip_template"],
                "usage_concept": concept,
                "template_image_description": desc,
            }
        )

    return pd.DataFrame(data)


def query_imkg_memes(graph: Graph) -> pd.DataFrame:
    query = """
    SELECT DISTINCT ?kym_src ?imgf_src  ?imgf_template_src ?about ?alt_text ?views ?upvotes ?id
    WHERE
    {
        ?kym_src <https://meme4.science/title> ?title;
                <https://meme4.science/about> ?about .
        ?imgf_src <https://imgflip.com/template_title> ?title; <https://imgflip.com/template> ?imgf_template_src .
        ?imgf_src <https://imgflip.com/view_count>  ?views; <https://imgflip.com/upvote_count> ?upvotes; <https://imgflip.com/alt_text> ?alt_text; <https://imgflip.com/templateId> ?id
    }
    """  # noqa E501

    data = []
    qres = graph.query(query)
    for row in track(qres, description="Querying IMKG memes..."):
        if match := re.search(
            r"^ *([^|]+) *\| *([^|]+) *\| *image tagged in *([^|]+) *\|[^|]*$",
            row.alt_text.strip(),
            re.IGNORECASE,
        ):
            instance_text = match.group(2).strip()

        data.append(
            {
                "kym_template": row.kym_src.toPython().strip(),
                "imgflip_template": row.imgf_template_src.toPython().strip(),
                "imgflip_template_id": row.id.toPython().strip(),
                "imgflip_instance": row.imgf_src.toPython().strip(),
                "usage": row.about.toPython().strip(),
                "text": instance_text,
                "upvotes": (0 if row.upvotes.toPython() == "NA" else int(row.upvotes.toPython().replace(",", ""))),
                "views": (0 if row.views.toPython() == "NA" else int(row.views.toPython().replace(",", ""))),
            }
        )

    return pd.DataFrame(data)


def generate_imkg_claims(
    imkg_memes_df: pd.DataFrame,
    imkg_memes_descriptions_df: pd.DataFrame,
    llm: Llm = None,
    group_size: int = 50,
    group_sort: Sequence[str] = ["views", "upvotes"],
) -> pd.DataFrame:
    generated_claims = []

    merged_df = pd.merge(imkg_memes_df, imkg_memes_descriptions_df, on="imgflip_template").reset_index()

    memes_df = (
        merged_df.sort_values(group_sort, ascending=False)
        .groupby(["kym_template", "imgflip_template"])
        .head(group_size)
    )

    for _, row in track(
        memes_df.iterrows(),
        total=len(memes_df),
        description="Generating IMKG claims...",
    ):
        tmpl = row["imgflip_template"].split("/")[-1]
        title = tmpl.replace("-", " ").strip()

        correction_claim = (
            llm.chat(
                messages=[
                    {
                        "role": "system",
                        "content": "You are a fact-checker and meme expert that provides short and acurate answers.",
                    },
                    {
                        "role": "user",
                        "content": f"Give a claim explaining the meaning of the following meme text. The text of the meme contains {row['usage_concept']} and humour.",  # noqa E501
                    },
                    {
                        "role": "user",
                        "content": f"The image used by the meme is \"{title}\". {row['usage']} {row['template_image_description']}",  # noqa E501
                    },
                    {
                        "role": "user",
                        "content": f"The text of the meme is: {row['text']}",
                    },
                    {
                        "role": "assistant",
                        "content": 'The one sentence claim explaning the meme is: "',
                    },
                ],
            )
            .split('"')[0]
            .strip()
        )

        claim = (
            llm.chat(
                messages=[
                    {
                        "role": "system",
                        "content": "You are a fact-checker and meme expert that provides short and acurate answers.",
                    },
                    {
                        "role": "user",
                        "content": "Give a claim that contradict the following meme explanation.",
                    },
                    {"role": "user", "content": correction_claim},
                    {
                        "role": "assistant",
                        "content": 'The one sentence contradictory claim is: "',
                    },
                ],
            )
            .split('"')[0]
            .strip()
        )

        vec1 = llm.embed(prompt=row["text"])
        vec2 = llm.embed(prompt=claim)
        vec3 = llm.embed(prompt=correction_claim)

        generated_claims.append(
            {
                "kym_template": row["kym_template"],
                "imgflip_template": row["imgflip_template"],
                "imgflip_template_id": row["imgflip_template_id"],
                "imgflip_instance": row["imgflip_instance"],
                "imgflip_instance_text": row["text"],
                "imgflip_instance_claim": claim,
                "imgflip_instance_correction_claim": correction_claim,
                "imgflip_instance_text_vec": vec1,
                "imgflip_instance_claim_vec": vec2,
                "imgflip_instance_correction_claim_vec": vec3,
            }
        )

    return pd.DataFrame(generated_claims)


def merge_claims_data(
    imkg_memes_df: pd.DataFrame,
    imkg_memes_descriptions_df: pd.DataFrame,
    imkg_claims_df: pd.DataFrame,
) -> pd.DataFrame:
    merged_claims_df = pd.merge(
        imkg_memes_df,
        imkg_memes_descriptions_df,
        on="imgflip_template",
    )
    merged_claims_df = pd.merge(
        merged_claims_df,
        imkg_claims_df,
        on=["imgflip_instance", "kym_template", "imgflip_template", "imgflip_template_id"],
    )

    merged_claims_df = merged_claims_df[
        [
            "kym_template",
            "imgflip_template",
            "imgflip_template_id",
            "imgflip_instance",
            "usage",
            "text",
            "upvotes",
            "views",
            "usage_concept",
            "template_image_description",
            "imgflip_instance_claim",
            "imgflip_instance_correction_claim",
        ]
    ]
    return merged_claims_df.rename(
        columns={
            "usage": "kym_template_usage",
            "text": "imgflip_instance_text",
            "upvotes": "imgflip_instance_upvotes",
            "views": "imgflip_instance_views",
            "usage_concept": "kym_template_usage_concept",
            "template_image_description": "imgflip_template_image_description",
            "claim": "imgflip_instance_claim",
            "claim_correction": "imgflip_instance_correction_claim",
        }
    )


def build_embeddings_database(
    merged_claims_df: pd.DataFrame,
    path: Path = DEFAULT_EMBEDDINGS_PATH,
    delete_collection: bool = False,
) -> chromadb.Collection:
    client = chromadb.PersistentClient(path=str(path))

    if delete_collection:
        client.delete_collection("claims")
    collection = client.get_or_create_collection("claims")

    collection.add(
        documents=merged_claims_df.imgflip_instance_claim.to_list(),
        metadatas=merged_claims_df[
            [
                "kym_template",
                "kym_template_usage",
                "kym_template_usage_concept",
                "imgflip_template_id",
                "imgflip_template",
                "imgflip_template_image_description",
                "imgflip_instance",
                "imgflip_instance_text",
                "imgflip_instance_claim",
                "imgflip_instance_correction_claim",
            ]
        ].to_dict("records"),
        ids=merged_claims_df.imgflip_instance.to_list(),
    )

    return collection
